import sys, optparse, os

class TransformToolContinuous:
	"""Point of the transformation function for tools which produce confidence score in form of continuous value."""
	point = None
	expAccuracy = None
	
	def __init__(self, argPoint, argExpAccuracy):
		self.point = argPoint
		self.expAccuracy = argExpAccuracy


class TransformToolCategorical:
	"""Point of the transformation function for tools which produce confidence score in form of categorical value."""
	category = None
	expAcc = None
	
	def __init__(self, argCategory, argExpAccuracy):
		self.category = argCategory
		self.expAccuracy = argExpAccuracy


class PredictSNP:
	"""Main class which is responsible for transformation between output of the individual tools and PredictSNP consensus."""
	dataFile = os.path.join(os.path.dirname(os.path.realpath(__file__)), "predictsnp.data")
	inputFile = None
	outputFile = None
	
	# Associative arrays for transformation of confidence score of individual tools to expected accuracies
	transform = {}
	transform["MAPP"] = []
	transform["nsSNPAnalyzer"] = []
	transform["PANTHER"] = []
	transform["PhD-SNP"] = []
	transform["PolyPhen-1"] = []
	transform["PolyPhen-2"] = []
	transform["SIFT"] = []
	transform["SNAP"] = []
	transform["PredictSNP"] = []
	
	# Associative arrays for thresholds of confidence score of individual tools (for considering mutations as deleterious / neutral)
	cutoff = {}
	cutoff["MAPP"] = 0.1
	cutoff["PolyPhen-2"] = 0.5
	cutoff["SIFT"] = 0.05
	cutoff["PredictSNP"] = 0.0
	
	def __init__(self, argInputFile, argOutputFile):
		"Get path to the input & output file."
		self.inputFile = argInputFile
		self.outputFile = argOutputFile
	
	def loadTransformationFunctions(self):
		"Load the transformation functions of the individual tools and PredictSNP consensus."
		
		if(os.path.isfile(self.dataFile) == False):  # check the existence of the file with transformation functions
			raise Exception("Data file with transformation functions (predictsnp.data) must be in the same folder as executable script predictsnp.py")
		
		actToolName = ""
		actToolType = ""
		for line in open(self.dataFile):
			if(len(line.strip()) == 0):          # Skip empty lines
				continue
			if(line.startswith("#") == True):    # Lines with character '#' divide the data file to the sections belonging to the individual tools & PredictSNP consensus
				actToolName = line[6 : line.find(";")].strip()
				actToolType = line[line.find(";")+1 : line.find('#',7)].strip()
				continue
			items = line.split("\t")
			if(actToolType.startswith("continuous")):  # parsing commands for tools which produce continous output (MAPP, PolyPhen-2, SIFT)
				point = float(items[0].strip())
				expAccuracy = float(items[1].strip())
				self.transform[actToolName].append(TransformToolContinuous(point, expAccuracy))
			else:                                      # parsing commands for tools which produce categoricalal output (PhD-SNP, PolyPhen-1, SNAP)
				category = items[0].replace(",", "").replace("-","").replace(" ","").upper().strip()   # remove all commas, dashes, white spaces ... (change of "case style" will not be critical for the parsing of users input)
				expAccuracy = float(items[1].strip())
				self.transform[actToolName].append(TransformToolCategorical(category, expAccuracy))
			
	
	def transformInputFile(self):
		"Transform all mutations from the input file by using PredictSNP consensus."
		
		# Open output file & write headline
		outfile = open(self.outputFile,"w")
		headline = "Identifier\tMutation\tPredictSNP prediction\tPredictSNP exp.acc.\tMAPP prediction\tMAPP exp.acc.\tPhD-SNP prediction\tPhD-SNP exp.acc.\tPolyPhen-1 prediction\tPolyPhen-1 exp.acc\tPolyPhen-2 prediction\tPolyPhen-2 exp.acc.\tSIFT prediction\tSIFT exp.acc.\tSNAP prediction\tSNAP exp.acc.\n"
		outfile.write(headline)
		
		if(os.path.isfile(self.inputFile) == False):  # check the existence of the input file with mutations for the evaluation
			raise Exception("Input file does not exists!")
		
		# Perform transformation (line by line)
		for line in open(self.inputFile,"r"):
			if((len(line.strip()) == 0) or (line.strip().startswith("#"))):  # Skip headline or empty lines
				continue
			      
			items = line.split("\t")  # parse line with definition of mutation and the output of the individual tools
			if(len(items) < 8):       # check format of the input file
				raise Exception("Input file is not in correct format! Please check that format of input file is the same as format of example file (sampleInput.txt).")
			
			identifier = items[0].strip()
			mutation = items[1].strip()
			
			neutrality = ""
			value = 0
			score = 0
			expAcc = 0
			numerator = 0
			denominator = 0
			outLine = ""
			
			# MAPP
			value = items[2].strip()
			if(value.startswith("?") == True):
				outLine += "unknown\tnull\t"
			else:
				score = float(value)
				expAcc = self.findExpAcc("MAPP", score)
				neutrality = "neutral" if (expAcc < 0) else "deleterious"
				outLine += "%s\t%.2f\t" % (neutrality, abs(expAcc))
				numerator += expAcc
				denominator += abs(expAcc)
			
			# PhD-SNP
			value = items[3].strip()
			if(value.startswith("?") == True):
				outLine += "unknown\tnull\t"
			else:
				expAcc = self.findExpAcc("PhD-SNP", value)
				neutrality = "neutral" if (expAcc < 0) else "deleterious"
				outLine += "%s\t%.2f\t" % (neutrality, abs(expAcc))
				numerator += expAcc
				denominator += abs(expAcc)
			
			# PolyPhen-1
			value = items[4].strip()
			if(value.startswith("?") == True):
				outLine += "unknown\tnull\t"
			else:
				expAcc = self.findExpAcc("PolyPhen-1", value)
				neutrality = "neutral" if (expAcc < 0) else "deleterious"
				outLine += "%s\t%.2f\t" % (neutrality, abs(expAcc))
				numerator += expAcc
				denominator += abs(expAcc)
			
			# PolyPhen-2
			value = items[5].strip()
			if(value.startswith("?") == True):
				outLine += "unknown\tnull\t"
			else:
				score = 1 - float(value)
				expAcc = self.findExpAcc("PolyPhen-2", score)
				neutrality = "neutral" if (expAcc < 0) else "deleterious"
				outLine += "%s\t%.2f\t" % (neutrality, abs(expAcc))
				numerator += expAcc
				denominator += abs(expAcc)
			
			# SIFT
			value = items[6].strip()
			if(value.startswith("?") == True):
				outLine += "unknown\tnull\t"
			else:
				value = items[6].strip().split(",")[1].strip()
				siftValueNeutrality = items[6].strip().split(",")[0].strip().upper()
				
				if(value.startswith("0.05") == True):
					score = 0.051 if siftValueNeutrality.startswith("TOLERATED") == True else 0.049
				else:
					score = float(value)
				expAcc = self.findExpAcc("SIFT", score)
				neutrality = "neutral" if (expAcc < 0) else "deleterious"
				outLine += "%s\t%.2f\t" % (neutrality, abs(expAcc))
				numerator += expAcc
				denominator += abs(expAcc)
			
			# PolyPhen-1
			value = items[7].strip()
			if(value.startswith("?") == True):
				outLine += "unknown\tnull\t"
			else:
				expAcc = self.findExpAcc("SNAP", value)
				neutrality = "neutral" if (expAcc < 0) else "deleterious"
				outLine += "%s\t%.2f\t" % (neutrality, abs(expAcc))
				numerator += expAcc
				denominator += abs(expAcc)
			
			predictSnpScore = numerator / denominator
			predictSnpExpAcc = self.findExpAcc("PredictSNP", predictSnpScore)
			
			predictSnpNeu = "deleterious" if predictSnpScore >= 0.0 else "neutral"
			predictSnpOut = "%s\t%.2f\t" % (predictSnpNeu, abs(predictSnpExpAcc))
			
			outLine = identifier + "\t" + mutation + "\t" + predictSnpOut + outLine + "\n"
			outfile.write(outLine)
		outfile.close()
	
	def findExpAcc(self, toolName, value):
		"Find the expected accuracy for specific value of confidence of specific tool."
		expAccuracy = 0
		if(toolName.startswith("MAPP") or toolName.startswith("PolyPhen-2") or toolName.startswith("SIFT") or toolName.startswith("PredictSNP")):       # tools which produce continuous value

			value = float(value)
			neutrality = "deleterious" if value <= self.cutoff[toolName] else "neutral"
			minDiff = 2.0
			i = 0
			for rec in self.transform[toolName]:
				actDiff = abs((rec.point+1) - (value+1))
				recNeutrality = "neutral" if rec.expAccuracy < 0 else "deleterious"
				if((actDiff < minDiff) and (neutrality.startswith(recNeutrality) == True)):
					minDiff = actDiff
					expAccuracy = rec.expAccuracy
				i += 1
			return expAccuracy

		elif(toolName.startswith("PhD-SNP") or toolName.startswith("PolyPhen-1") or toolName.startswith("SNAP")):  # tool which produce categorical value
			value = value.replace(",", "").replace("-","").replace(" ","").upper().strip()   # remove all commas, dashes, white spaces ... (change of "case style" will not be critical for the parsing of users input)
			for rec in self.transform[toolName]:
				if((value.startswith(rec.category) == True) and (rec.category.startswith(value) == True)):
					return rec.expAccuracy


def main():
	# Process command line arguments
	parser = optparse.OptionParser("usage: %prog [options] arg1 arg2")
	parser.add_option("-i", "--input", dest="inputFile", type="string", help="input file (confidence scores of underlying tools of PredictSNP)")
	parser.add_option("-o", "--output", dest="outputFile", type="string", help="output file")
	(options, args) = parser.parse_args()
	if len(sys.argv) != 5:
		parser.error("incorrect number of arguments")

	predict = PredictSNP(options.inputFile, options.outputFile)
	predict.loadTransformationFunctions()  # load transformation functions
	predict.transformInputFile()           # load mutations for the evaluation by the PredictSNP consensus


if __name__ == "__main__":
	main()
